"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.pipe = void 0;
function pipe(input) {
    var _a;
    var operations = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        operations[_i - 1] = arguments[_i];
    }
    var output = input;
    var lazyOperations = operations.map(function (op) {
        return "lazy" in op ? toPipedLazy(op) : undefined;
    });
    var operationIndex = 0;
    while (operationIndex < operations.length) {
        var lazyOperation = lazyOperations[operationIndex];
        if (lazyOperation === undefined || !isIterable(output)) {
            var operation = operations[operationIndex];
            output = operation(output);
            operationIndex++;
            continue;
        }
        var lazySequence = [];
        for (var j = operationIndex; j < operations.length; j++) {
            var lazyOp = lazyOperations[j];
            if (lazyOp === undefined) {
                break;
            }
            lazySequence.push(lazyOp);
            if (lazyOp.isSingle) {
                break;
            }
        }
        var accumulator = [];
        var iterator = output[Symbol.iterator]();
        while (true) {
            var result = iterator.next();
            if ((_a = result.done) !== null && _a !== void 0 ? _a : false) {
                break;
            }
            var shouldExitEarly = _processItem(result.value, accumulator, lazySequence);
            if (shouldExitEarly) {
                break;
            }
        }
        var isSingle = lazySequence[lazySequence.length - 1].isSingle;
        if (isSingle) {
            output = accumulator[0];
        }
        else {
            output = accumulator;
        }
        operationIndex += lazySequence.length;
    }
    return output;
}
exports.pipe = pipe;
function _processItem(item, accumulator, lazySequence) {
    if (lazySequence.length === 0) {
        accumulator.push(item);
        return false;
    }
    var lazyResult = { done: false, hasNext: false };
    var isDone = false;
    for (var i = 0; i < lazySequence.length; i++) {
        var lazyFn = lazySequence[i];
        var isIndexed = lazyFn.isIndexed, index = lazyFn.index, items = lazyFn.items;
        items.push(item);
        lazyResult = isIndexed ? lazyFn(item, index, items) : lazyFn(item);
        lazyFn.index++;
        if (lazyResult.hasNext) {
            if (lazyResult.hasMany) {
                var nextValues = lazyResult.next;
                for (var _i = 0, nextValues_1 = nextValues; _i < nextValues_1.length; _i++) {
                    var subItem = nextValues_1[_i];
                    var subResult = _processItem(subItem, accumulator, lazySequence.slice(i + 1));
                    if (subResult) {
                        return true;
                    }
                }
                return false;
            }
            item = lazyResult.next;
        }
        if (!lazyResult.hasNext) {
            break;
        }
        if (lazyResult.done) {
            isDone = true;
        }
    }
    if (lazyResult.hasNext) {
        accumulator.push(item);
    }
    if (isDone) {
        return true;
    }
    return false;
}
function toPipedLazy(op) {
    var lazy = op.lazy, lazyArgs = op.lazyArgs;
    var fn = lazy.apply(void 0, (lazyArgs !== null && lazyArgs !== void 0 ? lazyArgs : []));
    return Object.assign(fn, {
        isIndexed: lazy.indexed,
        isSingle: lazy.single,
        index: 0,
        items: [],
    });
}
function isIterable(something) {
    return (typeof something === "string" ||
        (typeof something === "object" &&
            something !== null &&
            Symbol.iterator in something));
}
